# -----
# Imports
# -----
from Models.L298N import L298N
from Models.ReadArduino import Arduino
from Models.sendEmail import SendEmail
import logging as log, time, os, MySQLdb
from threading import Thread
from queue import Queue
from RPi import GPIO

# -----
# Global imports
# -----
queue = Queue()
log.basicConfig(level=log.DEBUG)


# -----
# RunDevice class
# -----
class RunDevice:
    def __init__(self):
        """
        Let the logger know I exist
        """
        GPIO.setmode(GPIO.BCM)
        self.__arduino = Arduino('/dev/ttyACM0', 9600)
        time.sleep(1)
        self.__motor = L298N(6, 5, 19, 13)
        self.__brush = L298N(22, 27, 17, 18)
        self.__hallsensor = 21
        self.ultrasonic, self.irsensor, self.irdetect, self.wheel_rotations, self.battery_analog, self.container_analog = [], [], False, 0, [], []
        self.__EMAILBODY = "<h1>Hello {0}</h1><p>This message was sent to you by your CleanBot device, to warn you that the device battery level is low.</p><p>The device has been shut down.</p><p><b>Could you plug in the power when you have the chance?</b></p><h2>Thank you!</h2>"
        self.RUNNING = False
        self.PAUSED = False
        self.__rotateleft, self.__rotateright = False, False
        log.info("RunDevice Threading object created")

    # -----
    # Private help functions
    # -----

    def __get_email_list(self):
        """
            Get all email addresses from users in the DB with "Receive emails" enabled
            :return: Email addresses
            """
        try:
            db = MySQLdb.connect(host="localhost",
                                 user="project1-admin",
                                 passwd="Password4AdminAccess",
                                 db="project1")
            cursor = db.cursor()
            cursor.execute(
                "SELECT us.email, us.userFirstname from emaillist as el join user as us on us.userID = el.userID;")
            data = cursor.fetchall()
            result = []
            for d in data:
                result.append([d[0], d[1]])
            cursor.close()
            return result
        except Exception as e:
            return e

    def __check_for_safety_shutdown(self):
        """
        Check if the battery analog value is greater then the low threshold
        :return: shutdown if device battery level is critical
        """
        if (sum(self.battery_analog) / len(self.battery_analog)) < 800:
            self.stop()
            maillist = self.__get_email_list()
            for contact in maillist:
                SendEmail(contact[0], "Your CleanBot is low on battery!", self.__EMAILBODY.format(contact[1]))
            time.sleep(3 / 4)
            os.system("sudo shutdown now")

    def __execute_sql(self, sql, params=None):
        """
        Execute a SQL command
        :param sql: The SQL
        :param params: Extra safe parameters
        """
        conn = MySQLdb.connect(host="localhost",
                               user="project1-sensor",
                               passwd="Password4SensorAccess",
                               db="project1")
        cursor = conn.cursor()
        try:
            cursor.execute(sql, params)
            conn.commit()
        except Exception as e:
            conn.rollback()
        cursor.close()
        conn.close()

    # -----
    # Threading functions
    # -----
    def move(self):
        """
        Move function for the device
        :return: log messages about what the device is doing
        """
        while self.RUNNING:
            self.ultrasonic = queue.get(block=True)  # block=true: do not continue executing code before data input
            if self.__rotateleft:
                self.__rotateleft = False
                self.__motor.rotate_left_side()
                time.sleep(1 / 2)
            elif self.__rotateright:
                self.__rotateright = False
                self.__motor.rotate_right_side()
                time.sleep(1 / 2)
            if not self.PAUSED:
                self.__brush.rotate_forward()
                self.__motor.rotate_forward()
                if len(self.ultrasonic) > 3:
                    if self.ultrasonic[0] <= 15 or self.ultrasonic[1] <= 15 or self.ultrasonic[2] <= 15:
                        self.__motor.stop_rotating_forward()
                        self.__motor.rotate_reverse()
                        log.info("Reverse")
                        time.sleep(3 / 4)
                        self.__motor.stop_rotating_reverse()
                        self.__motor.rotate_right_side()
                        log.info("Turn right")
                        time.sleep(1 / 2)
                        self.__motor.rotate_forward()
                        log.info("Running forward")
            else:
                self.__brush.stop_rotating_forward()
                self.__motor.stop_rotating_forward()
        log.debug("Stopped running move")

    def get_data(self):
        """
        Get the data from the Arduino from serial connection
        :return: put the data list in the Threading queue
        """
        while self.RUNNING:
            self.ultrasonic = self.__arduino.get_data_array()
            try:
                self.battery_analog.append(self.ultrasonic[3])
                self.container_analog.append(self.ultrasonic[4])
            except Exception as e:
                log.warning("couldnt get data from battery: {}".format(e))
            queue.put(self.ultrasonic)
        log.debug("Stopped running get_data")

    def data_handler(self):
        """
        Upload the wheel rotations from the hall sensor callback to the database
        """
        while self.RUNNING:
            if self.wheel_rotations >= 5:
                self.__execute_sql(
                    "INSERT INTO devicerunsensors(sensorID, runID, sensorData) values (1,(select runID from devicerun where endTime is null),%s)",
                    (self.wheel_rotations,))
                self.wheel_rotations = 0
                log.info("Uploaded wheel rotations to db")
            if len(self.container_analog) >= 500:
                container_avg = sum(self.container_analog) / len(self.container_analog)
                self.__execute_sql(
                    "INSERT INTO devicerunsensors(sensorID, runID, sensorData) values (3,(select runID from devicerun where endTime is null),%s)",
                    (container_avg,))
                self.container_analog = []
                log.info("Uploaded container value to db")
            if len(self.battery_analog) >= 500:
                battery_avg = sum(self.battery_analog) / len(self.battery_analog)
                self.__execute_sql(
                    "INSERT INTO devicerunsensors(sensorID, runID, sensorData) values (2,(select runID from devicerun where endTime is null),%s)",
                    (battery_avg,))
                self.__check_for_safety_shutdown()
                self.battery_analog = []
                log.info("Uploaded battery value to db")
        log.debug("Stopped running data_handler")

    # -----
    # Callback functions
    # -----
    def hall_cb(self, p):
        """
        Callback function for a hall sensor
        :param p: the pin where the callback was called
        :return: increment the rotations of the wheel and upload to db at a certain value
        """
        self.wheel_rotations += 1
        log.info("wheel rotations: {}".format(self.wheel_rotations))

    # -----
    # Device control functions
    # -----

    def start(self):
        """
        Start the device automaticly
        """
        GPIO.setup(self.__hallsensor, GPIO.IN)
        GPIO.add_event_detect(self.__hallsensor, GPIO.FALLING, bouncetime=300, callback=self.hall_cb)
        log.info("Started threading")
        self.RUNNING, self.PAUSED = True, False

        move_thread = Thread(target=self.move)
        data_get_thread = Thread(target=self.get_data)
        data_handler_thread = Thread(target=self.data_handler)

        move_thread.start()
        data_get_thread.start()
        data_handler_thread.start()

    def stop(self):
        """
        Terminate the threads and halt the device
        """
        log.info("Stopped threading")
        try:
            GPIO.remove_event_detect(self.__hallsensor)
        except:
            log.warning("No hallsensor callback added yet")
        self.RUNNING = False
        GPIO.cleanup()

    def rotate_left(self):
        """
        Rotate the device to the left side, callable from the manual control page
        """
        if self.RUNNING:
            self.__rotateleft = True

    def rotate_right(self):
        """
        Rotate the device to the right side, callable from the manual control page
        """
        if self.RUNNING:
            self.__rotateright = True

    def halt_device(self):
        """
        Pause the device untill resume is pressed
        """
        if self.RUNNING:
            self.PAUSED = True

    def resume_device(self):
        """
        Continue device after pause mode
        """
        if self.RUNNING:
            self.PAUSED = False
