# -----
# Class imports
# -----
from flask import Flask, render_template, request, session, redirect
from os import urandom
from passlib.hash import argon2
from flaskext.mysql import MySQL
from Models.sendEmail import SendEmail
import datetime, os, logging as log
from math import pi as pi
from Models.RunDevice import RunDevice

# -----
# Class instances & configuration
# -----
app = Flask(__name__)
app.secret_key = urandom(32)
mysql = MySQL(app)
log.basicConfig(level=log.DEBUG)

# -----
# Database configuration
# -----
app.config['MYSQL_DATABASE_USER'] = 'project1-web'
app.config['MYSQL_DATABASE_PASSWORD'] = 'Password4WebAccess'
app.config['MYSQL_DATABASE_DB'] = 'project1'
app.config['MYSQL_DATABASE_HOST'] = 'localhost'

# -----
# Global variables
# -----
WELCOMEMAIL = "<h1>Welcome to CleanBot {0} {1}!</h1><p>To Register your account, please visit the CleanBot webinterface and go to the register screen. There you can enter your password, and get started!</p>"
DEVICE_WIDTH = 20
WHEEL_DIAMETRE = 6.5
WHEEL_SURFACE = pi * WHEEL_DIAMETRE

# -----
# Variables
# -----
device = RunDevice()


# -----
# Help functions
# -----
def get_data(sql, params=None):
    """
    Get data from the database
    :param sql: SQL command as string
    :param params: Extra parameters
    :return: The data collected by the SQL request
    """
    conn = mysql.connect()
    cursor = conn.cursor()
    try:
        cursor.execute(sql, params)
    except Exception as e:
        return e
    result = cursor.fetchall()
    db_data = [list(row) for row in result]
    cursor.close()
    conn.close()
    return db_data


def execute_sql(sql, params=None):
    """
    Execute a SQL command
    :param sql: The SQL
    :param params: Extra safe parameters
    """
    conn = mysql.connect()
    cursor = conn.cursor()
    try:
        cursor.execute(sql, params)
        conn.commit()
    except Exception as e:
        log.error(e)
        conn.rollback()
    cursor.close()
    conn.close()


def is_user_logged_in():
    """
    Check if a user is logged in with a session
    :return: True or False
    """
    if 'loggedin' and 'usermail' in session:
        return True
    return False


def email_in_db(email):
    """
    Checks if a certain email exists in the database
    :param email: the email to check
    :return: True or False
    """
    d = get_data("select count(*) from user where BINARY email = %s;", email)
    if d[0][0] == 1:
        return True
    return False


def check_password(email, password):
    """
    Check if the given password equals the database password
    :param email: The email to check the password from
    :param password: The password submitted by a user
    :return: True or False
    """
    d = get_data("SELECT password FROM user WHERE BINARY email = %s;", email)
    print("WW: {}".format(d[0][0]))
    if len(d[0][0]) > 1:
        if argon2.verify(password, d[0][0]):
            return True
    return False


def is_user_admin(email):
    """
    Check if a user is an administrator or not
    :param email: The user email stored in session variables
    :return: True or False
    """
    d = get_data("SELECT roleID FROM user WHERE binary email = %s;", email)
    if d[0][0] == 2:
        return True
    return False


def user_has_password(email):
    """
    Check if a user has a filled in password in the database
    :param email: the user email to be checked
    :return: True or False
    """
    d = get_data("SELECT password from user where binary email = %s;", email)
    if d[0][0] is None:
        return False
    return True


def is_user_disabled(email):
    """
    Check if a user is disabled by email
    :param email: the user email to be checked
    :return: True or False
    """
    d = get_data("SELECT disabled from user where email = %s;", email)
    if d[0][0] == 1:
        return True
    return False


def is_device_running():
    """
    Check if there is a device run without end date, if so, the device is running.
    :return: True or False
    """
    d = get_data("SELECT * FROM devicerun where startTime is not null and endTime is null;")
    if len(d) < 1:
        return False
    return True


# -----
# Routing functions
# -----
@app.route('/')
def index():
    """
    Index page
    :return: The home page if a user is logged in, otherwise redirecting to the login page
    """
    if is_user_logged_in():
        n = get_data("SELECT userFirstname from user where binary email = %s;", session['usermail'])
        battery = get_data("SELECT sensorData from devicerunsensors where sensorID = 2 order by dataTime desc limit 1;")
        bat = round((battery[0][0] * 0.41) - 320.99)
        container = get_data(
            "SELECT sensorData from devicerunsensors where sensorID = 3 order by dataTime desc limit 1;")
        return render_template('index.html', Pagename='Home', Name=n[0][0], Running=is_device_running(),
                               Batteryperc=bat, Containerlvl=container[0][0])
    return redirect('/login')


@app.route('/settings')
def settings():
    """
    Main settings page
    :return: The settings menu if user is logged in, else redirect to the login page
    """
    if is_user_logged_in():
        email = session['usermail']
        role = get_data("SELECT roleID FROM user WHERE binary email = %s;", email)
        return render_template('settings.html', Pagename='Settings', Role=role[0][0])
    return redirect('/login')


@app.route('/faq')
def faq():
    """
    Show the FAQ page
    :return: The FAQ page if a user is logged in, else redirect to the login page
    """
    if is_user_logged_in():
        return render_template('faq.html', Pagename='FAQ')
    return redirect('/')


@app.route('/login', methods=['GET', 'POST'])
def login():
    """
    Show the login screen
    :return: The login screen if no user is logged in, else redirect to home
    """
    if is_user_logged_in():
        return redirect('/')
    if request.method == 'POST':
        pwd = request.form.get('password')
        email = request.form.get('email')
        if email_in_db(email):
            if not is_user_disabled(email):
                if user_has_password(email):
                    if check_password(email, pwd):
                        session['loggedin'] = 1
                        session['usermail'] = email
                        return redirect('/')
        return render_template('login.html', Error=True)
    return render_template('login.html')


@app.route('/logout')
def logout():
    """
    Log the user out
    :return: The logout page message if user is logged in, else redirect to login page.
    """
    if is_user_logged_in():
        session.pop('loggedin')
        session.pop('usermail')
        return render_template('logout.html')
    return redirect('/login')


@app.route('/register', methods=['GET', 'POST'])
def register():
    """
    Registration page
    :return: the registration page template if no user is logged in, else redirect to the Index page
    """
    if not is_user_logged_in():
        if request.method == 'POST':
            email = request.form.get('email')
            pwd = request.form.get('password')
            pwd_repeat = request.form.get('passwordRepeat')
            if pwd == pwd_repeat:
                if len(pwd) > 5:
                    if email_in_db(email):
                        if not user_has_password(email):
                            execute_sql("UPDATE user SET password = %s WHERE binary email = %s;",
                                        (argon2.hash(pwd), email))
                            uid = get_data("SELECT userID from user WHERE binary email = %s;", email)
                            execute_sql("INSERT INTO emaillist (userID) VALUES (%s);", uid[0][0])
                            return redirect('/')
                        return render_template('register.html', RegisterError=True)
                    return render_template('register.html', ErrorEmail=True)
                return render_template('register.html', LengthError=True)
            return render_template('register.html', ErrorPwd=True)
        return render_template('register.html')
    return redirect('/')


@app.route('/statistics')
def statistics():
    """
    The statistics page with ChartJS charts
    :return: The statistics page if nobody is logged in, else redirect to the login page
    """
    if is_user_logged_in():
        dataset1 = get_data(
            "select date(startTime), round(SUM(TIME_TO_SEC(endTime) - TIME_TO_SEC(startTime)) /60 ) from devicerun where startTime and endTime BETWEEN (NOW() - INTERVAL 7 DAY) AND NOW() group by month(startTime), day(startTime);")
        dataset2 = get_data(
            "select round(((sum(sensorData) * %s) * %s) / 1000), date(dataTime) from devicerunsensors where sensorID = 1 group by date(dataTime) order by dataTime asc limit 7;",
            (WHEEL_SURFACE, DEVICE_WIDTH))
        print(dataset2)
        surface_done = [d[0] for d in dataset2]
        surface_days = [d[1].day for d in dataset2]
        surface_month = [d[1].month for d in dataset2]
        surface_year = [d[1].year for d in dataset2]
        days = [d[0].day for d in dataset1]
        months = [d[0].month for d in dataset1]
        years = [d[0].year for d in dataset1]
        values = [d[1] for d in dataset1]
        return render_template("statistics.html", Pagename="Statistics", Days=days, Months=months, Years=years,
                               Values=values, Surface=surface_done, SurfaceDays=surface_days,
                               SurfaceMonth=surface_month, SurfaceYear=surface_year)
    return redirect('/')


@app.route('/startauto')
def startauto():
    """
    Start device autonomous
    :return: The device starts running automaticly & user is redirected to home, else if no person is logged in,
    redirect to login page
    """
    if is_user_logged_in():
        if not device.RUNNING:
            global device
            device = RunDevice()
            device.start()
        userid = get_data("SELECT userID from user where binary email = %s", session['usermail'])
        execute_sql("INSERT INTO devicerun(userID, startTime) VALUES(%s, %s);", (userid[0][0], datetime.datetime.now()))
    return redirect('/')


@app.route('/startmanual')
def startmanual():
    if is_user_logged_in():
        if not device.RUNNING:
            global device
            device = RunDevice()
            device.start()
            userid = get_data("SELECT userID from user where binary email = %s", session['usermail'])
            execute_sql("INSERT INTO devicerun(userID, startTime) VALUES(%s, %s);",
                        (userid[0][0], datetime.datetime.now()))
        return render_template('manualcontrol.html', Pagename='Control')
    return redirect('/')


@app.route('/stopcleaning')
def stopcleaning():
    """
    Stop the device from cleaning
    :return: the home page if a user is logged in, else redirect login page
    """
    if is_user_logged_in():
        if device.RUNNING:
            device.stop()
        record = get_data("SELECT runID from devicerun where endTime is null;")
        execute_sql("UPDATE devicerun SET endTime = %s WHERE runID = %s;", (datetime.datetime.now(), record[0][0]))
    return redirect('/')


@app.route('/shutdown')
def shutdown():
    """
    Shut the device down by sending a terminal command
    :return: shut the device down if a user is logged in, else go to the login page
    """
    if is_user_logged_in():
        if is_device_running():
            record = get_data("SELECT runID from devicerun where endTime is null;")
            execute_sql("UPDATE devicerun SET endTime = %s WHERE runID = %s;", (datetime.datetime.now(), record[0][0]))
        os.system('sudo shutdown now')
    return redirect('/')


@app.route('/manualcontrol')
def manualcontrol():
    """
    :return: Show the page with the manual control buttons if logged in, else redirect to login
    """
    if is_user_logged_in():
        if device.RUNNING:
            return render_template('manualcontrol.html', Pagename='Control', State=device.PAUSED)
    return redirect('/')


# -----
# Manual control routes
# -----
@app.route('/rotateleft')
def rotateleft():
    """
    Call the RunDevice function to rotate to the left
    :return: the manual control page
    """
    if is_user_logged_in():
        if device.RUNNING:
            device.rotate_left()
            return redirect('/manualcontrol')
    return redirect('/')


@app.route('/rotateright')
def rotateright():
    """
    Call the RunDevice function to rotate to the left
    :return: the manual control page
    """
    if is_user_logged_in():
        if device.RUNNING:
            device.rotate_right()
            return redirect('/manualcontrol')
    return redirect('/')


@app.route('/halt')
def halt():
    """
    Call the RunDevice function to halt the device
    :return: the manual control page
    """
    if is_user_logged_in():
        if device.RUNNING:
            device.halt_device()
            return redirect('/manualcontrol')
    return redirect('/')


@app.route('/continuerun')
def continuerun():
    """
    Call the RunDevice function to continue the device
    :return: the manual control page
    """
    if is_user_logged_in():
        if device.RUNNING:
            device.resume_device()
            return redirect('/manualcontrol')
    return redirect('/')


# -----
# Routing for admins only
# -----
@app.route('/manageusers')
def manageusers():
    """
    Shows the screen with all users, admins and possibility to add new users
    """
    if is_user_logged_in():
        users = get_data("SELECT userID, userFirstname, userLastname FROM user where roleID = 1 and disabled = 0;")
        admins = get_data("SELECT userID, userFirstname, userLastname FROM user where roleID = 2 and disabled = 0;")
        roles = get_data("SELECT roleID,roleName from userroles;")
        return render_template("manageusers.html", Pagename='Users', Users=users, Admins=admins, Roles=roles)
    return redirect('/login')


@app.route('/wipepwd/<userid>')
def wipepwd(userid):
    """
    Wipe a user's password so it can be resetted
    :param userid: UserID to wipe password from
    """
    if is_user_logged_in():
        if is_user_admin(session['usermail']):
            execute_sql("UPDATE user SET password=NULL WHERE userID = %s;", userid)
            user = get_data("SELECT email from user where userID = %s;", userid)
            SendEmail(user[0][0], "Your password has ben reset", "<p>Go to register to create a new password</p>")
            return redirect('/manageusers')
    return redirect('/')


@app.route('/removeusr/<userid>')
def removeusr(userid):
    """
    Fully remove a user from the database
    :param userid: UserID to delete data from
    """
    if is_user_logged_in():
        if is_user_admin(session['usermail']):
            execute_sql("DELETE FROM emaillist where userID = %s;", userid)
            execute_sql("UPDATE user SET disabled = 1 WHERE userID = %s;", userid)
            print("executed")
            return redirect('/manageusers')
    return redirect('/')


@app.route('/adduser', methods=['GET', 'POST'])
def adduser():
    """
    Create a new user, and send an email to confirm the address
    """
    if is_user_logged_in():
        if is_user_admin(session['usermail']):
            roles = get_data("SELECT roleID, roleName from userroles;")
            if request.method == 'POST':
                email = request.form.get('email')
                fname = request.form.get('fname')
                lname = request.form.get('lname')
                role = request.form.get('role')
                if not email_in_db(email):
                    try:
                        execute_sql(
                            "INSERT INTO user(email, userFirstname, userLastname, roleID) VALUES(%s, %s, %s, %s);",
                            (email, fname, lname, role))
                    except:
                        return render_template('adduser.html', Roles=roles, Error=True)
                    return redirect('/manageusers')
                return render_template('adduser.html', Roles=roles, Regerror=True)
            return render_template('adduser.html', Roles=roles)
    return redirect('/')


@app.route('/showdisabled')
def showdisabled():
    """
    Show every disabled user from the database
    :return: A page containing every disabled user, if the user is logged in as admin, else redirect to the login
    or home page.
    """
    if is_user_logged_in():
        if is_user_admin(session['usermail']):
            d = get_data("SELECT userID, userFirstname, userLastname from user where disabled = 1;")
            print(d)
            return render_template('disabled.html', Pagename="Users", Data=d)
    return redirect('/')


@app.route('/enableusr/<userid>')
def enableusr(userid):
    """
    Re-enable a user that was disabled
    :param userid: The user id to be re-enabled
    :return: the page of disabled users if you are a logged in administrator, else the login or home page
    """
    if is_user_logged_in():
        if is_user_admin(session['usermail']):
            execute_sql("UPDATE user SET disabled=0 WHERE userID = %s;", userid)
            return redirect('/showdisabled')
    return redirect('/')


# -----
# Setting pages
# -----
@app.route('/emailsettings', methods=['GET', 'POST'])
def emailsettings():
    """
    Shows the email settings page
    :return: The email page with togglebutton according to database values
    """
    if is_user_logged_in():
        if request.method == 'POST':
            val = request.form.get('preference')
            if val == "1":
                execute_sql("INSERT INTO emaillist (userID) VALUES ((select userID from user where email = %s));",
                            session['usermail'])
            elif val == "0":
                execute_sql("DELETE FROM emaillist WHERE userID = (select userID from user where email = %s);",
                            session['usermail'])
            return redirect('/settings')

        ev = get_data(
            "SELECT count(e.userID) from emaillist as e join user as u on u.userID = e.userID where u.email = %s;",
            session['usermail'])
        return render_template('emailsettings.html', Pagename="Settings", EmailValue=ev[0][0])
    return redirect('/')


# -----
# Error handling
# -----
@app.errorhandler(404)
def page_not_found(e):
    """1
    Catch any 404 error with an error page
    :param e: the error text
    :return: If a user is logged in: the 404 page, else redirect to the login screen
    """
    if is_user_logged_in():
        return render_template('404.html', Pagename="404")
    return redirect('/')


# -----
# Run the application
# -----
if __name__ == '__main__':
    app.run(host='0.0.0.0', debug=True)
